@@ -1,33 +1,21 @@
 #include "core.h"
 #include "llvm-c/Target.h"
-#include "llvm/Support/Host.h"
 #include "llvm-c/TargetMachine.h"
-#include "llvm/Target/TargetMachine.h"
-#include "llvm/IR/LegacyPassManager.h"
 #include "llvm/Analysis/TargetLibraryInfo.h"
-#include "llvm/ADT/Triple.h"
-#include "llvm/Support/TargetRegistry.h"
+#include "llvm/IR/LegacyPassManager.h"
 #include "llvm/IR/Type.h"
+#include "llvm/MC/TargetRegistry.h"
+#include "llvm/Target/TargetMachine.h"
+#include "llvm/TargetParser/Host.h"
+#include "llvm/TargetParser/Triple.h"
 
 #include <cstdio>
 #include <cstring>
 #include <sstream>
 
-
 namespace llvm {
 
-
-inline LLVMTargetLibraryInfoRef wrap(TargetLibraryInfo *TLI) {
-    return reinterpret_cast<LLVMTargetLibraryInfoRef>(TLI);
-}
-
-inline TargetLibraryInfo *unwrap(LLVMTargetLibraryInfoRef TLI) {
-    return reinterpret_cast<TargetLibraryInfo*>(TLI);
-}
-
-inline Target *unwrap(LLVMTargetRef T) {
-    return reinterpret_cast<Target*>(T);
-}
+inline Target *unwrap(LLVMTargetRef T) { return reinterpret_cast<Target *>(T); }
 
 inline TargetMachine *unwrap(LLVMTargetMachineRef TM) {
     return reinterpret_cast<TargetMachine *>(TM);
@@ -37,7 +25,7 @@ inline LLVMTargetMachineRef wrap(TargetMachine *TM) {
     return reinterpret_cast<LLVMTargetMachineRef>(TM);
 }
 
-}
+} // namespace llvm
 
 extern "C" {
 
@@ -46,21 +34,40 @@ LLVMPY_GetProcessTriple(const char **Out) {
     *Out = LLVMPY_CreateString(llvm::sys::getProcessTriple().c_str());
 }
 
+API_EXPORT(void)
+LLVMPY_GetTripleParts(const char *triple_str, const char **arch_out,
+                      const char **vendor_out, const char **os_out,
+                      const char **environment_out) {
+    // Normalize the triple string
+    auto triple_str_norm = llvm::Triple::normalize(triple_str);
+    auto triple = llvm::Triple(triple_str_norm);
+
+    *arch_out = LLVMPY_CreateString(
+        llvm::Triple::getArchTypeName(triple.getArch()).data());
+    *vendor_out = LLVMPY_CreateString(
+        llvm::Triple::getVendorTypeName(triple.getVendor()).data());
+    *os_out =
+        LLVMPY_CreateString(llvm::Triple::getOSTypeName(triple.getOS()).data());
+    *environment_out = LLVMPY_CreateString(
+        llvm::Triple::getEnvironmentTypeName(triple.getEnvironment()).data());
+}
+
 /**
  * Output the feature string to the output argument.
  * Features are prefixed with '+' or '-' for enabled or disabled, respectively.
  * Features are separated by ','.
  */
 API_EXPORT(int)
-LLVMPY_GetHostCPUFeatures(const char **Out){
-    llvm::StringMap<bool> features;
+LLVMPY_GetHostCPUFeatures(const char **Out) {
+    // https://github.com/llvm/llvm-project/pull/97824
+    llvm::StringMap<bool> features = llvm::sys::getHostCPUFeatures();
     std::ostringstream buf;
-    if ( llvm::sys::getHostCPUFeatures(features) ) {
+    if (!features.empty()) {
         for (auto &F : features) {
-            if (buf.tellp()){
+            if (buf.tellp()) {
                 buf << ',';
             }
-            buf << ((F.second? "+": "-") + F.first()).str();
+            buf << ((F.second ? "+" : "-") + F.first()).str();
         }
         *Out = LLVMPY_CreateString(buf.str().c_str());
         return 1;
@@ -79,69 +86,43 @@ LLVMPY_GetHostCPUName(const char **Out) {
 }
 
 API_EXPORT(int)
-LLVMPY_GetTripleObjectFormat(const char *tripleStr)
-{
+LLVMPY_GetTripleObjectFormat(const char *tripleStr) {
     return llvm::Triple(tripleStr).getObjectFormat();
 }
 
 API_EXPORT(LLVMTargetDataRef)
-LLVMPY_CreateTargetData(const char *StringRep)
-{
+LLVMPY_CreateTargetData(const char *StringRep) {
     return LLVMCreateTargetData(StringRep);
 }
 
-
-//// Nothing is creating a TargetLibraryInfo
-//    void
-//    LLVMPY_AddTargetLibraryInfo(LLVMTargetLibraryInfoRef TLI,
-//                                LLVMPassManagerRef PM)
-//    {
-//        LLVMAddTargetLibraryInfo(TLI, PM);
-//    }
-
 API_EXPORT(void)
-LLVMPY_CopyStringRepOfTargetData(LLVMTargetDataRef TD, char** Out)
-{
+LLVMPY_CopyStringRepOfTargetData(LLVMTargetDataRef TD, char **Out) {
     *Out = LLVMCopyStringRepOfTargetData(TD);
 }
 
 API_EXPORT(void)
-LLVMPY_DisposeTargetData(LLVMTargetDataRef TD)
-{
-    LLVMDisposeTargetData(TD);
-}
-
+LLVMPY_DisposeTargetData(LLVMTargetDataRef TD) { LLVMDisposeTargetData(TD); }
 
 API_EXPORT(long long)
-LLVMPY_ABISizeOfType(LLVMTargetDataRef TD, LLVMTypeRef Ty)
-{
-    return (long long) LLVMABISizeOfType(TD, Ty);
+LLVMPY_ABISizeOfType(LLVMTargetDataRef TD, LLVMTypeRef Ty) {
+    return (long long)LLVMABISizeOfType(TD, Ty);
 }
 
 API_EXPORT(long long)
-LLVMPY_ABISizeOfElementType(LLVMTargetDataRef TD, LLVMTypeRef Ty)
-{
+LLVMPY_OffsetOfElement(LLVMTargetDataRef TD, LLVMTypeRef Ty, int Element) {
     llvm::Type *tp = llvm::unwrap(Ty);
-    if (!tp->isPointerTy())
+    if (!tp->isStructTy())
         return -1;
-    tp = tp->getPointerElementType();
-    return (long long) LLVMABISizeOfType(TD, llvm::wrap(tp));
+    return (long long)LLVMOffsetOfElement(TD, Ty, Element);
 }
 
 API_EXPORT(long long)
-LLVMPY_ABIAlignmentOfElementType(LLVMTargetDataRef TD, LLVMTypeRef Ty)
-{
-    llvm::Type *tp = llvm::unwrap(Ty);
-    if (!tp->isPointerTy())
-        return -1;
-    tp = tp->getPointerElementType();
-    return (long long) LLVMABIAlignmentOfType(TD, llvm::wrap(tp));
+LLVMPY_ABIAlignmentOfType(LLVMTargetDataRef TD, LLVMTypeRef Ty) {
+    return (long long)LLVMABIAlignmentOfType(TD, Ty);
 }
 
-
 API_EXPORT(LLVMTargetRef)
-LLVMPY_GetTargetFromTriple(const char *Triple, const char **ErrOut)
-{
+LLVMPY_GetTargetFromTriple(const char *Triple, const char **ErrOut) {
     char *ErrorMessage;
     LLVMTargetRef T;
     if (LLVMGetTargetFromTriple(Triple, &T, &ErrorMessage)) {
@@ -153,43 +134,35 @@ LLVMPY_GetTargetFromTriple(const char *Triple, const char **ErrOut)
 }
 
 API_EXPORT(const char *)
-LLVMPY_GetTargetName(LLVMTargetRef T)
-{
-    return LLVMGetTargetName(T);
-}
+LLVMPY_GetTargetName(LLVMTargetRef T) { return LLVMGetTargetName(T); }
 
 API_EXPORT(const char *)
-LLVMPY_GetTargetDescription(LLVMTargetRef T)
-{
+LLVMPY_GetTargetDescription(LLVMTargetRef T) {
     return LLVMGetTargetDescription(T);
 }
 
 API_EXPORT(LLVMTargetMachineRef)
-LLVMPY_CreateTargetMachine(LLVMTargetRef T,
-                           const char *Triple,
-                           const char *CPU,
-                           const char *Features,
-                           int         OptLevel,
-                           const char *RelocModel,
-                           const char *CodeModel,
-                           int         EmitJITDebug,
-                           int         PrintMC)
-{
+LLVMPY_CreateTargetMachine(LLVMTargetRef T, const char *Triple, const char *CPU,
+                           const char *Features, int OptLevel,
+                           const char *RelocModel, const char *CodeModel,
+                           int PrintMC, int JIT, const char *ABIName) {
     using namespace llvm;
-    CodeGenOpt::Level cgol;
-    switch(OptLevel) {
+
+    // https://github.com/llvm/llvm-project/pull/66295
+    CodeGenOptLevel cgol;
+    switch (OptLevel) {
     case 0:
-        cgol = CodeGenOpt::None;
+        cgol = CodeGenOptLevel::None;
         break;
     case 1:
-        cgol = CodeGenOpt::Less;
+        cgol = CodeGenOptLevel::Less;
         break;
     case 3:
-        cgol = CodeGenOpt::Aggressive;
+        cgol = CodeGenOptLevel::Aggressive;
         break;
     case 2:
     default:
-        cgol = CodeGenOpt::Default;
+        cgol = CodeGenOptLevel::Default;
     }
 
     CodeModel::Model cm;
@@ -213,7 +186,8 @@ LLVMPY_CreateTargetMachine(LLVMTargetRef T,
             cm = CodeModel::Large;
     }
 
-    Optional<Reloc::Model> rm;
+    // llvm::Optional removed in llvm17
+    std::optional<Reloc::Model> rm;
     std::string rms(RelocModel);
     if (rms == "static")
         rm = Reloc::Static;
@@ -223,49 +197,42 @@ LLVMPY_CreateTargetMachine(LLVMTargetRef T,
         rm = Reloc::DynamicNoPIC;
 
     TargetOptions opt;
-//     opt.JITEmitDebugInfo = EmitJITDebug;
-    opt.PrintMachineCode = PrintMC;
+    opt.MCOptions.ShowMCInst = PrintMC;
+    opt.MCOptions.ABIName = ABIName;
 
-    return wrap(unwrap(T)->createTargetMachine(Triple, CPU, Features, opt,
-                                               rm, cm, cgol));
+    bool jit = JIT;
+
+    return wrap(unwrap(T)->createTargetMachine(Triple, CPU, Features, opt, rm,
+                                               cm, cgol, jit));
 }
 
-
 API_EXPORT(void)
-LLVMPY_DisposeTargetMachine(LLVMTargetMachineRef TM)
-{
+LLVMPY_DisposeTargetMachine(LLVMTargetMachineRef TM) {
     return LLVMDisposeTargetMachine(TM);
 }
 
 API_EXPORT(void)
-LLVMPY_GetTargetMachineTriple(LLVMTargetMachineRef TM, const char **Out)
-{
+LLVMPY_GetTargetMachineTriple(LLVMTargetMachineRef TM, const char **Out) {
     // result is already strdup()ed by LLVMGetTargetMachineTriple
     *Out = LLVMGetTargetMachineTriple(TM);
 }
 
 API_EXPORT(void)
-LLVMPY_SetTargetMachineAsmVerbosity(LLVMTargetMachineRef TM, int verbose)
-{
+LLVMPY_SetTargetMachineAsmVerbosity(LLVMTargetMachineRef TM, int verbose) {
     LLVMSetTargetMachineAsmVerbosity(TM, verbose);
 }
 
 API_EXPORT(LLVMMemoryBufferRef)
-LLVMPY_TargetMachineEmitToMemory (
-    LLVMTargetMachineRef TM,
-    LLVMModuleRef M,
-    int use_object,
-    const char ** ErrOut
-    )
-{
+LLVMPY_TargetMachineEmitToMemory(LLVMTargetMachineRef TM, LLVMModuleRef M,
+                                 int use_object, const char **ErrOut) {
     LLVMCodeGenFileType filetype = LLVMAssemblyFile;
-    if (use_object) filetype = LLVMObjectFile;
+    if (use_object)
+        filetype = LLVMObjectFile;
 
     char *ErrorMessage;
     LLVMMemoryBufferRef BufOut;
     int err = LLVMTargetMachineEmitToMemoryBuffer(TM, M, filetype,
-                                                  &ErrorMessage,
-                                                  &BufOut);
+                                                  &ErrorMessage, &BufOut);
     if (err) {
         *ErrOut = LLVMPY_CreateString(ErrorMessage);
         LLVMDisposeMessage(ErrorMessage);
@@ -276,60 +243,31 @@ LLVMPY_TargetMachineEmitToMemory (
 }
 
 API_EXPORT(LLVMTargetDataRef)
-LLVMPY_CreateTargetMachineData(LLVMTargetMachineRef TM)
-{
-    return llvm::wrap(new llvm::DataLayout(llvm::unwrap(TM)->createDataLayout()));
+LLVMPY_CreateTargetMachineData(LLVMTargetMachineRef TM) {
+    return llvm::wrap(
+        new llvm::DataLayout(llvm::unwrap(TM)->createDataLayout()));
 }
 
 API_EXPORT(void)
-LLVMPY_AddAnalysisPasses(
-    LLVMTargetMachineRef TM,
-    LLVMPassManagerRef PM
-    )
-{
+LLVMPY_AddAnalysisPasses(LLVMTargetMachineRef TM, LLVMPassManagerRef PM) {
     LLVMAddAnalysisPasses(TM, PM);
 }
 
-
-API_EXPORT(const void*)
-LLVMPY_GetBufferStart(LLVMMemoryBufferRef MB)
-{
-    return LLVMGetBufferStart(MB);
-}
+API_EXPORT(const void *)
+LLVMPY_GetBufferStart(LLVMMemoryBufferRef MB) { return LLVMGetBufferStart(MB); }
 
 API_EXPORT(size_t)
-LLVMPY_GetBufferSize(LLVMMemoryBufferRef MB)
-{
-    return LLVMGetBufferSize(MB);
-}
+LLVMPY_GetBufferSize(LLVMMemoryBufferRef MB) { return LLVMGetBufferSize(MB); }
 
 API_EXPORT(void)
-LLVMPY_DisposeMemoryBuffer(LLVMMemoryBufferRef MB)
-{
+LLVMPY_DisposeMemoryBuffer(LLVMMemoryBufferRef MB) {
     return LLVMDisposeMemoryBuffer(MB);
 }
 
-API_EXPORT(int)
-LLVMPY_HasSVMLSupport(void)
-{
-#ifdef HAVE_SVML
-    return 1;
-#else
-    return 0;
-#endif
+API_EXPORT(void)
+LLVMPY_AddTargetLibraryInfoPass(LLVMPassManagerRef PM, const char *TripleStr) {
+    using namespace llvm;
+    unwrap(PM)->add(new TargetLibraryInfoWrapperPass(Triple(TripleStr)));
 }
 
-
-/*
-
-If needed:
-
-explicit TargetLibraryInfoWrapperPass(const Triple &T);
-
-void LLVMAddTargetLibraryInfo(LLVMTargetLibraryInfoRef TLI,
-                              LLVMPassManagerRef PM) {
-  unwrap(PM)->add(new TargetLibraryInfoWrapperPass(*unwrap(TLI)));
-}
-*/
-
 } // end extern "C"
